; Single 16-Segment LED Display Ornament
;========================================

;Display a string on a single 16 segment LED display by displaying each character
;for about 1/2s followed by a short blanking of the display. Idea stolen from:
;http://www.evilmadscientist.com/article.php/ledholidaycode

; ATTiny2313 @ 500 kHz (internal R/C)
; Fuses:  low = 0x42 (internal R/C 4MHz, divide Clock by 8, BOD enabled)
;        high = 0x9D (BOD voltage 1.8V, preserve EEPROM through chip erase)

;DISPLAY SEGMENTS:
;      a1    a2
;    ===== =====
;   | \   |   / |
; f |  \h |i /j | b
;   |   \ | /   |
;   | g1 \|/ g2 |
;    ===== =====
;   |    /|\    |
; e |   / | \   | c
;   | k/ l| m\  |
;   | /   |   \ |
;    ===== =====
;     d1     d2

;PORTS (display type: VQB 201 C):
;  PB0 = d1
;  PB1 = k
;  PB2 = e / RS232 Rx
;  PB3 = f / RS232 GND
;  PB4 = h
;  PB5 = i
;  PB6 = a1
;  PB7 = a2
;  PD0 = (unused, connected to Display Anode and +3V)
;  PD1 = j
;  PD2 = g1
;  PD3 = c
;  PD4 = m
;  PD5 = l
;  PD6 = d2
;  PA0 = g2
;  PA1 = b

;EEPROM:
;  0x00      - character display duration (* 10 ms)
;  0x01      - pause between characters (* 10 ms)
;  0x02-0x41 - string (max. 64 byte)

;VERSION HISTORY:
;  Version 0.3  2006-12-17  Arne Rossius
;  Version 0.9  2009-12-19  Arne Rossius

.include "../include/tn2313def.inc"

.def	char_delay = R14
.def	pause_delay = R15

.def	temp = R16
.def	temp2 = R17
.def	anz1 = R18
.def	anz2 = R19
.def	w1 = R20
.def	w2 = R21

.equ	CharDelay = 40 ;display duration for each character (unit: ~10 ms)
.equ	PauseDelay = 5 ;short blanking duration between characters (unit: ~10ms)

;===============================================================================

.dseg
	RAM_String: .byte 64
	
	RAM_Stack: .byte 16

;===============================================================================

.cseg
.org 0x000
	rjmp	start
	
charset:
	;ASCII charset for 16 segment LED display
	.include "vqb201c-charset.asm"
	
start:
	;init stack pointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;disable analog comparator
	ldi	temp, 1<<ACD
	out	ACSR, temp
	
	;init ports
	ldi	temp, 0xFF
	out	DDRB, temp
	out	PORTB, temp
	ldi	temp, 0x7A
	out	DDRD, temp
	ldi	temp, 0x76
	out	PORTD, temp
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0x07
	out	PORTA, temp
	
	;check which mode is selected
	ldi	temp, 10
	rcall	wait10ms
	sbic	PIND, 2 ;PD2 low in spite of pullup -> RS232 connected
	rjmp	play
	
;===============================================================================
	
;--------------------
;   "TRANSFER" mode
;--------------------
	
transfer:
	;init registers
	ldi	XL, LOW(RAM_String)
	ldi	XH, HIGH(RAM_String)
	
getc:
	;wait for transmission to begin
	sbis	PIND, 2
	rjmp	getc
	
	;receive byte
	rcall	waitbit ;startbit
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit
	
	rcall	waitbit ;data 0
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit
	
	rcall	waitbit ;data 1
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 2
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 3
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 4
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 5
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 6
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 7
	clc
	sbis	PIND, 2
	sec
	ror	temp
	brcs	getc_starterr ;start bit was not 1
	rcall	waitbit

	rcall	waitbit
	sbic	PIND, 2
	rjmp	getc_stoperr ;stop bit is not 0
	
	;save byte to RAM buffer
	cpi	XL, RAM_String+64 ;prevent buffer overflow
	brlo	PC+2
	ldi	XL, RAM_String+63
	st	X+, temp

	;evaluate byte
	cpi	temp, 0x03 ;CTRL-C -> go to play mode
	breq	play
	cpi	temp, 0x04 ;CTRL-D -> evaluate string
	breq	getc_done
	cpi	temp, 0x0a ;LF -> evaluate string
	breq	getc_done
	cpi	temp, 0x0d ;CR -> evaluate string
	breq	getc_done
	cpi	temp, 0x08 ;BACKSP -> remove last character and the backspace character from buffer
	breq	getc_backsp
	rjmp	getc_end
	
getc_backsp:
	cpi	XL, RAM_String+2
	brlo	getc_end
	subi	XL, 2
	
getc_starterr:
getc_stoperr:
getc_end:
	;wait until line is idle (needed if stop bit is not 0) and receive next byte
	sbic	PIND, 2
	rjmp	getc_end
	rjmp	getc
	
;--------------------
	
getc_done:
	;terminate string
	dec	XL
	ldi	temp, 0x00 ;terminate string
	st	X, temp	
getc_eval:
	;evaluate received string
	cpi	XL, RAM_String ;string is empty (contains only end marker)
	brne	PC+2
	rjmp	transfer
	ldi	XL, RAM_String
	ld	temp, X
	cpi	temp, 0x1b ;ESC
	breq	getc_control
	
	;save received string to EEPROM
	ldi	YL, 0x02
getc_savestr_loop:
	ld	temp, X+
	rcall	eeprom_write
	inc	YL
	cpi	temp, 0x00
	brne	getc_savestr_loop
	rjmp	transfer
	
getc_control:
	;control string received (begins with ESC)
	inc	XL
	ld	temp, X+
	cpi	temp, 'd' ;char display duration
	breq	getc_delay
	cpi	temp, 'D'
	breq	getc_delay
	cpi	temp, 'p' ;pause between chars
	breq	getc_pause
	cpi	temp, 'P'
	breq	getc_pause
	cpi	temp, 'q' ;quit -> switch to display mode
	breq	play
	cpi	temp, 0x1b ;ESC -> switch to display mode
	breq	play
	rjmp	transfer
	
getc_delay:
	rcall	dec2bin
	ldi	YL, 0x00
	rcall	eeprom_write
	rjmp	transfer

getc_pause:
	rcall	dec2bin
	ldi	YL, 0x01
	rcall	eeprom_write
	rjmp	transfer
	
;===============================================================================

;--------------------
;   "PLAY" mode
;--------------------
	
play:
	ldi	temp, 0x7E
	out	DDRD, temp
	out	PORTD, temp
	
	;get constants from EEPROM
	ldi	YL, 0x00
	rcall	eeprom_read
	mov	char_delay, temp
	ldi	YL, 0x01
	rcall	eeprom_read
	mov	pause_delay, temp
	
play_start:
	ldi	YL, 0x02
	
play_loop:
	;load next character
	rcall	eeprom_read
	inc	YL
	cpi	temp, 0
	breq	play_start ;restart string
	lsl	temp ;convert to 16-segment character
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm	anz1, Z+
	lpm	anz2, Z+

	;display character
	out	PORTB, anz1
	mov	temp, anz2
	lsl	temp
	andi	temp, 0x7E
	out	PORTD, temp
	mov	temp, anz2
	swap	temp
	lsr	temp
	lsr	temp
	andi	temp, 0x03
	out	PORTA, temp
	mov	temp, char_delay
	rcall	wait10ms
	
	;blank display
	ldi	temp, 0xFF
	out	PORTB, temp
	ldi	temp, 0x7F
	out	PORTD, temp
	ldi	temp, 0x03
	out	PORTA, temp
	mov	temp, pause_delay
	rcall	wait10ms
	
	rjmp	play_loop

;===============================================================================

;--------------------
;   subroutines
;--------------------

dec2bin:
	;convert 3-digit decimal number from RAM[X,X+2] to 1 byte binary
	ld	temp2, X+ ;first digit: multiply by 100 = (64 + 32 + 4)
	andi	temp2, 0x03
	lsl	temp2 ;temp2 = temp2 * 2
	mov	temp, temp2
	swap	temp ;temp = (temp2*2) * 16 = temp2 * 32
	lsl	temp2 ;temp2 = (temp2*2) * 2 = temp2 * 4
	add	temp, temp2
	swap	temp2 ;temp2 = ((temp2*2)*2) * 16 = temp2 * 64
	add	temp, temp2
	
	ld	temp2, X+ ;second digit: multiply by 10 = (8 + 2)
	andi	temp2, 0x0F
	lsl	temp2 ;temp2 = temp2 * 2
	add	temp, temp2
	lsl	temp2 ;temp2 = (temp2*2) * 4 = temp2 * 8
	lsl	temp2
	add	temp, temp2
	
	ld	temp2, X+ ;first digit
	andi	temp2, 0x0F
	add	temp, temp2
	
	ret
	
;--------------------

wait10ms:
	;wait for temp * 10 ms
	ldi	w1, 7
	clr	w2
wait10ms_loop:
	dec	w2
	brne	wait10ms_loop
	dec	w1
	brne	wait10ms_loop
	dec	temp
	brne	wait10ms
	ret

;--------------------

waitbit:
	;wait 1/2 rs232 bit length (203 cycles + rcall + 2 cycles for saving the bit)
	ldi	w1, 66
waitbit_loop:
	dec	w1
	brne	waitbit_loop
	nop
	ret

;--------------------

eeprom_write:
	sbic	EECR, EEPE
	rjmp	eeprom_write
	out	EEAR, YL
	out	EEDR, temp
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
	ret

;--------------------

eeprom_read:
	sbic	EECR, EEPE
	rjmp	eeprom_read
	out	EEAR, YL
	sbi	EECR, EERE
	in	temp, EEDR
	ret
